# !/usr/bin/env python
# encoding: utf-8
"""
experiments.py - Implements the experiments about the different parameters
~ Daniel Cortild, 11 November 2023
"""

# External Imports
import numpy as np
import matplotlib.pyplot as plt
from tqdm.auto import tqdm

# Internal Imports
from .Image import Image
from .Mask import Mask
from .Algorithm import Algorithm

# Running Algoroithm
def runAlgorithm(rho, lamb, sigma, percent, tolerance, max_it, method):
    """
    Runs the algorithm given different parameters.
    Input:
        rho         Step size parameter
        lamb        Regularisation parameter
        sigma       Relaxation parameter
        percent     Percentage of erased pixels
        tolerance   Stopping tolerance
        max_it      Maximum number of iterations
        method      Acceleration method selected
    Output:
        its         Iterations required to run the algorithm
        time        Time required to run the algorithm
    """
    image = Image(name="Image", dims=(512, 512))
    image.load_image("Venice.jpeg")
    mask = Mask(percentage=percent, dims=(512, 512))
    image_corrupt = mask.mask_image(image.image)
    _, _, _, _, its, time, _ = Algorithm(image_corrupt, mask, rho, lamb, sigma, 
                                         tolerance, max_it, method).run(False, False)
        
    return its, time

# Error Plots
def plotItsTime(parameters, max_it, its_S, its_H, its_N, its_R, 
                  time_S, time_H, time_N, time_R, xlabel, title):
    """
    Plots the results of an experiment given the results.
    Input:
        parameters      The list of paramaters computed against
        max_it          Maximum number of iterations
        its_S           List of iterations required for static
        its_H           List of iterations required for heavy-ball
        its_N           List of iterations required for Nesterov
        its_R           List of iterations required for reflected
        time_S          List of times required for static
        time_H          List of times required for heavy-ball
        time_N          List of times required for Nesterov
        time_R          List of times required for reflected
        xlabel          The X-label for the plot
        title           The plot title
    Output:
        None
    """
    # Post analyse the data
    max_time = max(max(time_S), max(time_H), max(time_N), max(time_R))
    time_S = [time_S[i] if its_S[i] < max_it else max_time for i in range(len(time_S))]
    time_H = [time_H[i] if its_H[i] < max_it else max_time for i in range(len(time_H))]
    time_N = [time_N[i] if its_N[i] < max_it else max_time for i in range(len(time_N))]
    time_R = [time_R[i] if its_R[i] < max_it else max_time for i in range(len(time_R))]
    
    fig, axs = plt.subplots(1, 2, figsize=(14, 3.5), dpi=600)

    axs[0].title.set_text("Iterations to reach $10^{-3}$ tolerance")
    axs[0].plot(parameters, its_S, label="Static")
    axs[0].plot(parameters, its_H, label="Heavy-Ball")
    axs[0].plot(parameters, its_N, label="Nesterov")
    axs[0].plot(parameters, its_R, label="Reflected")
    axs[0].axhline(y=max_it, color="r", label="Did Not Converge", linestyle="--")
    axs[0].set_ylabel("Iterations")
    axs[0].set_xlabel(xlabel)
    axs[0].legend()

    axs[1].title.set_text("Time to reach $10^{-3}$ tolerance")
    axs[1].plot(parameters, time_S, label="Static")
    axs[1].plot(parameters, time_H, label="Heavy-Ball")
    axs[1].plot(parameters, time_N, label="Nesterov")
    axs[1].plot(parameters, time_R, label="Reflected")
    axs[1].axhline(y=max_time, color="r", label="Did Not Converge", linestyle="--")
    axs[1].set_ylabel("Time in seconds")
    axs[1].set_xlabel(xlabel)
    axs[1].legend()

    plt.show()

def plotExperiments(rho, sigma, lamb, percent):
    """
    Plots the results of an experiment based on different parameters.
    Input:
        rho         Step size parameter (Float or list of floats)
        sigma       Regularisation parameter (Float or list of floats)
        lamb        Relaxation parameter (Float or list of floats)
        percent     Percentage of erased pixels (Float or list of floats)
    Output:
        its_S       List of iterations required for static
        its_H       List of iterations required for heavy-ball
        its_N       List of iterations required for Nesterov
        its_R       List of iterations required for reflected
        time_S      List of times required for static
        time_H      List of times required for heavy-ball
        time_N      List of times required for Nesterov
        time_R      List of times required for reflected
    """
    # Import the Image
    image = Image(name="Image", dims=(512, 512))
    image.load_image("Venice.jpeg")

    # Universal Parameters
    tolerance = 1e-3
    max_it = 100

    # Check parameters
    if sum(map(lambda x: type(x) in [list, np.ndarray], [rho, sigma, lamb, percent])) != 1 and \
       sum(map(lambda x: type(x) in [float, int], [rho, sigma, lamb, percent])) != 3:
        raise ValueError("Only one input should be a list, others should be floats")

    # Get number of tests
    if type(rho) in [list, np.ndarray]:
        N = len(rho)
        rhos = rho
        sigmas = [sigma] * N
        lambs = [lamb] * N
        percents = [percent] * N
        xAxis = r"Step size ($\rho$)"
        title = "Analysis on the step size"
    elif type(sigma) in [list, np.ndarray]:
        N = len(sigma)
        rhos = [rho] * N
        sigmas = sigma
        lambs = [lamb] * N
        percents = [percent] * N
        xAxis = r"Regularisation parameter ($\sigma$)"
        title = "Analysis on the regularisation parameter"
    elif type(lamb) in [list, np.ndarray]:
        N = len(lamb)
        rhos = [rho] * N
        sigmas = [sigma] * N
        lambs = lamb
        percents = [percent] * N
        xAxis = r"Relaxation parameter ($\lambda$)"
        title = "Analysis on the relaxation parameter"
    elif type(percent) in [list, np.ndarray]:
        N = len(percent)
        rhos = [rho] * N
        sigmas = [sigma] * N
        lambs = [lamb] * N
        percents = percent
        xAxis = "Percentage of erased pixels"
        title = "Analysis on the percentage of erased pixels"
    else:
        raise ValueError("At least one input should be a list")

    # Set empty lists for results
    its_S   = [0] * N
    its_H   = [0] * N
    its_N   = [0] * N
    its_R   = [0] * N
    time_S  = [0] * N
    time_H  = [0] * N
    time_N  = [0] * N
    time_R  = [0] * N

    # Run the Algorithm
    for i, (rho, sigma, lamb, percent) in enumerate(tqdm(zip(rhos, sigmas, lambs, percents))):
        its_S[i], time_S[i] = runAlgorithm(rho, lamb, sigma, percent,
                                  tolerance, max_it, method="static")
        its_H[i], time_H[i] = runAlgorithm(rho, lamb, sigma, percent,
                                    tolerance, max_it, method="heavyball")
        its_N[i], time_N[i] = runAlgorithm(rho, lamb, sigma, percent,
                                    tolerance, max_it, method="nesterov")
        its_R[i], time_R[i] = runAlgorithm(rho, lamb, sigma, percent,
                                    tolerance, max_it, method="reflected")

    # Plot the results
    plotItsTime(parameters, max_it, its_S, its_H, its_N, its_R, 
                time_S, time_H, time_N, time_R, xAxis, title)

    return its_S, its_H, its_N, its_R, time_S, time_H, time_N, time_R

def plotExperimentRegularisation(rho, sigmas, lamb, percent, method):
    """
    Plots different images for different regularisation parameters.
    Input:
        rho         Step size selected
        sigmas      Relaxation parameters selected
        lamb        Regularisation parameter selected
        percent     Percentage of erased pixels selected 
        method      Acceleration method chosen
    Output:
        None
    """
    # Check input
    if (not type(rho) in [float, int]) or (not type(lamb) in [float, int]) or (not type(percent) in [float, int]) or \
        type(sigmas) not in [list, np.ndarray] or method not in ["static", "heavyball", "nesterov", "reflected"]:
        raise ValueError("Input is incorrect.")
    if len(sigmas) != 6:
        raise ValueError("List `sigmas` should contain 6 elements exactly.")

    # Universal Parameters
    tolerance = 1e-3
    max_it = 100

    # Import the Image
    image = Image(name="Image", dims=(512, 512))
    image.load_image("Venice.jpeg")
    mask = Mask(percentage=percent, dims=(512, 512))
    image_corrupt = mask.mask_image(image.image)

    # Create Algorithm Instances
    sols = [None] * len(sigmas)
    for i, sigma in enumerate(tqdm(sigmas)):
        sols[i] = Algorithm(image_corrupt, mask, rho, lamb, sigma, 
                            tolerance, max_it, method=method).run()[0]

    # Plot Original and Result
    fig, axs = plt.subplots(2, 4, figsize=(12, 6), dpi=100)

    axs[0,0].title.set_text("Original Image")
    axs[0,0].imshow(image.image, vmin=0, vmax=1)
    axs[0,0].set_axis_off()
    axs[0,0].set_facecolor("white")

    axs[0,1].title.set_text("Corrupt Image")
    axs[0,1].imshow(image_corrupt, vmin=0, vmax=1)
    axs[0,1].set_axis_off()
    axs[0,1].set_facecolor("white")

    axs[0,2].title.set_text(r"Recovered with $\sigma=0.0625$")
    axs[0,2].imshow(sols[0], vmin=0, vmax=1)
    axs[0,2].set_axis_off()
    axs[0,2].set_facecolor("white")

    axs[0,3].title.set_text(r"Recovered with $\sigma=0.25$")
    axs[0,3].imshow(sols[1], vmin=0, vmax=1)
    axs[0,3].set_axis_off()
    axs[0,3].set_facecolor("white")

    axs[1,0].title.set_text(r"Recovered with $\sigma=0.5$")
    axs[1,0].imshow(sols[2], vmin=0, vmax=1)
    axs[1,0].set_axis_off()
    axs[1,0].set_facecolor("white")

    axs[1,1].title.set_text(r"Recovered with $\sigma=1$")
    axs[1,1].imshow(sols[3], vmin=0, vmax=1)
    axs[1,1].set_axis_off()
    axs[1,1].set_facecolor("white")

    axs[1,2].title.set_text(r"Recovered with $\sigma=4$")
    axs[1,2].imshow(sols[4], vmin=0, vmax=1)
    axs[1,2].set_axis_off()
    axs[1,2].set_facecolor("white")

    axs[1,3].title.set_text(r"Recovered with $\sigma=16$")
    axs[1,3].imshow(sols[5], vmin=0, vmax=1)
    axs[1,3].set_axis_off()
    axs[1,3].set_facecolor("white")

    plt.show()


#!/usr/bin/env python
# encoding: utf-8
"""
Image.py - Implements an Image Class
~ Daniel Cortild, 21 March 2023
"""

# External Imports
from PIL import Image as ImagePIL
import numpy as np


class Image:
    """
    Provides a Class for a Masked Image, consisting of an image, a mask, and operations
    to modify the image and mask other images according to the same mask.
    Parameters:
        name                  Name of image, located in the same directory as code file
        dims                  Size to which the image should be resize
    Public Methods:
        load_image            Loads the image in the appropriate format
    Private Methods:
        
    """
    
    def __init__(self, name = "Image", dims = (256, 256)):
        self.name = name
        self.dims = dims

    def load_image(self, file_name) -> None:
        """ @public
        Loads the image
        """
        self.image = ImagePIL.open(file_name)
        self.image = self.image.resize(self.dims)
        self.image = np.asarray(self.image, dtype=np.float64) / 255



#!/usr/bin/env python
# encoding: utf-8
"""
__init__.py - __init__ file for the inpainter module
~ Daniel Cortild, 21 March 2023
"""

# from .Image import Image
# from .Mask import Mask
# from .Algorithm import Algorithm
from .experiments import plotExperiments, plotItsTime, plotExperimentRegularisation
from .getInpainted import getInpainted


# !/usr/bin/env python
# encoding: utf-8
"""
getInpainted.py - Returns the inpainted images
~ Daniel Cortild, 11 November 2023
"""

# External Imports
import numpy as np
import matplotlib.pyplot as plt

# Internal Imports
from .Image import Image
from .Mask import Mask
from .Algorithm import Algorithm


def getInpainted(rho, sigma, lamb, percent):
    """
    Returns the inpainted images.
    Input:
        rho         Step size
        sigma       Regularisation parameter
        lamb        Relaxation parameter
        percent     Percentage of erased pixels
    Output:
        None
    """
    # Import the Image
    image = Image(name="Image", dims=(512, 512))
    image.load_image("Venice.jpeg")
    mask = Mask(percentage=0.5, dims=(512, 512))
    image_corrupt = mask.mask_image(image.image)

    # Set the parameters
    tolerance = 1e-3
    max_it = 100

    # Create Algorithm Instance
    sol_S, hist_S, hist_T_S, hist_F_S, its_S, time_S, F_S = Algorithm(
                                        image_corrupt, mask, rho, lamb, sigma, 
                                        tolerance, max_it, method="static").run()
    sol_H, hist_H, hist_T_H, hist_F_H, its_H, time_H, F_H = Algorithm(
                                        image_corrupt, mask, rho, lamb, sigma, 
                                        tolerance, max_it, method="heavyball").run()
    sol_N, hist_N, hist_T_N, hist_F_N, its_N, time_N, F_N = Algorithm(
                                        image_corrupt, mask, rho, lamb, sigma, 
                                        tolerance, max_it, method="nesterov").run()
    sol_R, hist_R, hist_T_R, hist_F_R, its_R, time_R, F_R = Algorithm(
                                        image_corrupt, mask, rho, lamb, sigma, 
                                        tolerance, max_it, method="reflected").run()

    # Plot Original and Result
    fig, axs = plt.subplots(2, 3, figsize=(10, 5), dpi=200)

    axs[0,0].title.set_text("Original Image")
    axs[0,0].imshow(image.image, vmin=0, vmax=1)
    axs[0,0].set_axis_off()
    axs[0,0].set_facecolor("white")

    axs[0,1].title.set_text("Corrupt Image")
    axs[0,1].imshow(image_corrupt, vmin=0, vmax=1)
    axs[0,1].set_axis_off()
    axs[0,1].set_facecolor("white")

    axs[0,2].title.set_text("Static")
    axs[0,2].imshow(sol_S, vmin=0, vmax=1)
    axs[0,2].set_axis_off()
    axs[0,2].set_facecolor("white")

    axs[1,0].title.set_text("Heavy Ball")
    axs[1,0].imshow(sol_H, vmin=0, vmax=1)
    axs[1,0].set_axis_off()
    axs[1,0].set_facecolor("white")

    axs[1,1].title.set_text("Nesterov")
    axs[1,1].imshow(sol_N, vmin=0, vmax=1)
    axs[1,1].set_axis_off()
    axs[1,1].set_facecolor("white")

    axs[1,2].title.set_text("Reflected")
    axs[1,2].imshow(sol_R, vmin=0, vmax=1)
    axs[1,2].set_axis_off()
    axs[1,2].set_facecolor("white")

    plt.show()

    # Error Plots
    fig, axs = plt.subplots(1, 4, figsize=(10, 2.2), dpi=200)
    plt.subplots_adjust(left=0.1,
                        bottom=0.1,
                        right=0.9,
                        top=0.9,
                        wspace=0.2,
                        hspace=0.45)
    plt.tight_layout()

    axs[0].title.set_text("$|X_k-X_{k-1}|^2$")
    l1 = axs[0].plot([np.linalg.norm(hist_S[i] - hist_S[i-1])**2 for i in range(1, len(hist_S))], 
                        label="Static")
    l2 = axs[0].plot([np.linalg.norm(hist_H[i] - hist_H[i-1])**2 for i in range(1, len(hist_H))], 
                        label="Heavy Ball")
    l3 = axs[0].plot([np.linalg.norm(hist_N[i] - hist_N[i-1])**2 for i in range(1, len(hist_N))], 
                        label="Nesterov")
    l4 = axs[0].plot([np.linalg.norm(hist_R[i] - hist_R[i-1])**2 for i in range(1, len(hist_R))], 
                        label="Reflected")
    axs[0].set_yscale('log')
    axs[0].set_xlabel("Iteration (k)")

    axs[1].title.set_text(r"$|T_kX_k-X_{k}|^2$")
    axs[1].plot([np.linalg.norm(hist_T_S[i] - hist_S[i])**2 for i in range(1, len(hist_S))], 
                label="Static")
    axs[1].plot([np.linalg.norm(hist_T_H[i] - hist_H[i])**2 for i in range(1, len(hist_H))], 
                label="Heavy Ball")
    axs[1].plot([np.linalg.norm(hist_T_N[i] - hist_N[i])**2 for i in range(1, len(hist_N))], 
                label="Nesterov")
    axs[1].plot([np.linalg.norm(hist_T_R[i] - hist_R[i])**2 for i in range(1, len(hist_R))], 
                label="Reflected")
    axs[1].set_yscale('log')
    axs[1].set_xlabel("Iteration (k)")

    axs[2].title.set_text(r"$|X_k-X^*|^2$")
    axs[2].plot([np.linalg.norm(hist_S[i] - image.image)**2 for i in range(1, len(hist_S))], 
                label="Static")
    axs[2].plot([np.linalg.norm(hist_H[i] - image.image)**2 for i in range(1, len(hist_H))], 
                label="Heavy Ball")
    axs[2].plot([np.linalg.norm(hist_N[i] - image.image)**2 for i in range(1, len(hist_N))], 
                label="Nesterov")
    axs[2].plot([np.linalg.norm(hist_R[i] - image.image)**2 for i in range(1, len(hist_R))], 
                label="Reflected")
    axs[2].set_yscale('log')
    axs[2].set_xlabel("Iteration (k)")

    axs[3].title.set_text(r"$|F(X_k)-X^*|^2$")
    axs[3].plot([np.linalg.norm(hist_F_S[i] - F_S(image.image))**2 for i in range(1, len(hist_S))], 
                label="Static")
    axs[3].plot([np.linalg.norm(hist_F_H[i] - F_H(image.image))**2 for i in range(1, len(hist_H))], 
                label="Heavy Ball")
    axs[3].plot([np.linalg.norm(hist_F_N[i] - F_N(image.image))**2 for i in range(1, len(hist_N))], 
                label="Nesterov")
    axs[3].plot([np.linalg.norm(hist_F_R[i] - F_R(image.image))**2 for i in range(1, len(hist_R))], 
                label="Reflected")
    axs[3].set_yscale('log')
    axs[3].set_xlabel("Iteration (k)")

    fig.legend([l1, l2, l3, l4], 
            labels=["Static", "Heavy-Ball", "Nesterov", "Reflected"],
                loc='lower center', 
                bbox_to_anchor=(0.5, -0.18),
                ncol=4)

    plt.show()


# !/usr/bin/env python
# encoding: utf-8
"""
Algorithm.py - Implements the Algorithm
~ Daniel Cortild, 21 March 2023
"""

# External imports
import numpy as np
import scipy as sp
import scipy.linalg
from scipy.optimize import fsolve
from tqdm.auto import trange                # type: ignore
import time


# The proximal operator of the nuclear norm
def svd_shrink(X, r):
    """
    Returns the proximal of the nuclear norm
    """
    try:
        U, S, VT = sp.linalg.svd(X, full_matrices=False)
        S_shrink = np.maximum(S - r, 0)
        X_shrink = ( U * S_shrink ) @ VT
        return X_shrink
    except np.linalg.LinAlgError as err:
        return X
    
def nuc_norm(X):
    return np.linalg.norm(fold(X, 0), ord='nuc') + np.linalg.norm(fold(X, 1), ord='nuc')

# Corresponds to (.)_(1) and (.)_(2)
def fold(X, axis):
    """
    Transforms a (N, M, 3) tensor to a (N, 3*M) or (3*N, M) tensor
    Reverses un fold
    """
    (a, b) = (1, 3) if axis == 0 else (3, 1)
    return X.reshape(X.shape[0] * a, X.shape[1] * b, order='F')

# Corresponds to iota_1 and iota_2
def unfold(X, axis):
    """
    Transforms a (N, 3*M) or (3*N, M) tensor to a (N, M, 3) tensor
    Reverses fold
    """
    (a, b) = (1, 3) if axis == 0 else (3, 1)
    return X.reshape(X.shape[0] // a, X.shape[1] // b, 3, order='F')


class Algorithm:
    """
    Solves minimisation problems over a Hilbert space H of the type:
            min_{x in H} f(x) + g(x) + h(Lx)
    Parameters:
        proxf               The proximal operator of f
        proxg               The proximal operator of g
        LgradhL             The operator L^*(grad_h(L))
        Z_init              Initial guess of Z
        lamb                Value of lambda in (0,1) [Default: 0.5]
        rho                 Value of rho in (0,2) [Default: 1]
        beta                The inverse of the Lipschitz constant of grad_h [Default: 1]
        alpha_static        Boolean expression whether alpha is static or not [Default: False]
    Public Methods:
        run                Runs the algorithm
    Private Methods:
        iterate            Runs a single iteration of the algorithm
        error              Compute error estimate used for stopping criterion
    """

    def __init__(self, image, mask, rho, lamb, sigma, tolerance, max_it, method):
        self.rho = rho
        self.lamb = lamb
        self.tolerance = tolerance
        self.max_it = max_it
        self.sigma = sigma
        
        self.R = lambda X1, X0: np.linalg.norm(X1 - X0) / np.linalg.norm(X0)
        self.A = lambda X: mask.mask_image(X)
        self.proxf = lambda X: unfold(svd_shrink(fold(X, 0), rho*sigma), 0)
        self.proxg = lambda X: unfold(svd_shrink(fold(X, 1), rho*sigma), 1)
        
        if method == "static":
            alpha = beta = 0
        elif method == "heavyball":
            alpha = fsolve(lambda x: lamb * x * (1+x) - (1/lamb - 1) - 1e-4, 0.5)
            beta = 0
        elif method == "nesterov":
            alpha = beta = fsolve(lambda x: x * (1 + x) + (1/lamb - 1) * x * (1-x) - (1/lamb - 1) * (1-x) - 1e-4, 0.5)
        elif method == "reflected":
            alpha = 0
            beta = fsolve(lambda x: (1-lamb) * x * (1+x) + (1/lamb - 1) * x * (1-x) - (1/lamb - 1) * (1-x) - 1e-4, 0.5)
        else:
            raise ValueError("No correct method specified")
        self.get_alpha = lambda k: (1 - 1 / (k+1)) * alpha
        self.get_beta  = lambda k: (1 - 1 / (k+1)) * beta
        
        self.X_corrupt = image.copy()
        self.F = lambda X: np.linalg.norm(X-self.X_corrupt)**2/2 + sigma * nuc_norm(X)

    def __iterate(self, X_previous, X_actual, k):
        """ @private
        Perform the iterations according to Algorithm 2
        """
        alpha    = self.get_alpha(k)
        beta     = self.get_beta(k)
        lambd    = self.lamb
        
        Y       = X_actual + alpha * (X_actual - X_previous)
        Z       = X_actual + beta * (X_actual - X_previous)
        X_g     = self.proxg(Z)
        X_T     = Z - X_g + self.proxf(2 * X_g - Z - self.rho * (self.A(X_g) - self.X_corrupt))
        X_next  = (1 - lambd) * Y + lambd * X_T
        
        return X_actual, X_next, X_T

    def run(self, bregman=False, verbose=True):
        """ @public
        Run the algorithm given the number of iterations and the iterator
        """
        X_previous = np.zeros_like(self.X_corrupt)
        X_next = self.X_corrupt
        X_hist = []
        X_T_hist = []
        X_F_hist = []
        its = 0

        start = time.time()
        
        for its in trange(self.max_it, disable = not verbose):
            X_previous, X_next, X_T = self.__iterate(X_previous, X_next, its)
            X_hist.append(X_previous)
            X_F_hist.append(self.F(X_previous))
            X_T_hist.append(X_T)
            if self.R(X_next, X_previous) < self.tolerance:
                break

        solution = self.proxg(X_next)
                
        return solution, X_hist, X_T_hist, X_F_hist, its + 1, time.time() - start, self.F



#!/usr/bin/env python
# encoding: utf-8
"""
Image.py - Implements a Mask Class
~ Daniel Cortild, 21 March 2023
"""

# External Imports
from .Image import Image
import numpy as np


class Mask:
    """
    Provides a Class for a Masked Image, consisting of an image, a mask, and operations
    to modify the image and mask other images according to the same mask.
    Parameters:
        name                  Name of image
        dims                  Size to which the image should be resize
    Public Methods:
        mask_image            Masks the image
    Private Methods:
        create_mask           Creates the mask
    """
    
    def __init__(self, percentage = 0.5, dims = (256, 256)):
        self.dims = dims
        self.__create_mask(percentage, dims)
        
    def __create_mask(self, percentage, dims) -> None:
        """ @private
        Creates the mask
        """
        total_pixels = np.prod(dims)
        mask_size = int(total_pixels * percentage)
        mask = (np.array([0] * mask_size + [1] * (total_pixels - mask_size)))
        np.random.shuffle(mask)
        self.mask = mask.reshape(dims)
        
    def mask_image(self, image):
        """ @public
        Mask the image
        """
        mask_3d: np.ndarray = np.repeat(self.mask[:, :, None], 3, axis=2)
        masked_image = np.multiply(image, mask_3d)
        return masked_image
